package com.flow.service.impl;


import com.baomidou.mybatisplus.core.toolkit.StringUtils;
import com.flow.common.core.exception.BaseException;
import com.flow.common.core.util.ApplicationContextUtil;
import com.flow.common.oauth2.utils.SecurityContextUtil;
import com.flow.dao.AuthDao;
import com.flow.entity.Menu;
import com.flow.entity.User;
import com.flow.model.*;
import com.flow.service.AuthService;
import com.wf.captcha.ArithmeticCaptcha;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.oauth2.common.OAuth2AccessToken;
import org.springframework.security.oauth2.provider.ClientDetails;
import org.springframework.security.oauth2.provider.ClientDetailsService;
import org.springframework.security.oauth2.provider.TokenRequest;
import org.springframework.security.oauth2.provider.password.ResourceOwnerPasswordTokenGranter;
import org.springframework.security.oauth2.provider.token.TokenStore;
import org.springframework.stereotype.Service;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

import javax.servlet.http.HttpServletRequest;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

@Service
public class AuthServiceImpl implements AuthService {
    @Autowired
    private ClientDetailsService clientDetailsService;
    @Autowired
    private TokenStore tokenStore;
    @Autowired
    private StringRedisTemplate redisTemplate;
    @Autowired
    private AuthDao authDao;

    @Override
    public ImageCode getImageCode() {
        // 算术验证码：三个参数分别为宽、高、位数
        ArithmeticCaptcha captcha = new ArithmeticCaptcha(111, 36, 2);
        // code保存到redis
        String code = captcha.text().toLowerCase();
        String uuid = UUID.randomUUID().toString();
        redisTemplate.opsForValue().set("image_code:" + uuid, code, 2L, TimeUnit.MINUTES);
        // 返回结果
        return new ImageCode(captcha.toBase64(), uuid);
    }

    @Override
    public AccessToken login(LoginForm loginForm) {
        String uuid = loginForm.getUuid();
        String code = loginForm.getCode();
        if (StringUtils.isBlank(code) || StringUtils.isBlank(uuid)) {
            throw new BaseException("验证码不能为空");
        }
        String redisCode = redisTemplate.opsForValue().get("image_code:" + uuid);
        if (StringUtils.isBlank(redisCode)) {
            throw new BaseException("验证码不存在或已过期");
        }
        if (!code.equalsIgnoreCase(redisCode)) {
            throw new BaseException("验证码错误");
        }
        HttpServletRequest request = ((ServletRequestAttributes) RequestContextHolder.getRequestAttributes()).getRequest();
        request.setAttribute("login_type", "social_login");
        ClientDetails clientDetails = clientDetailsService.loadClientByClientId("client");
        Map<String, String> requestParameters = new HashMap<>(5);
        requestParameters.put("grant_type", "password");
        requestParameters.put("username", loginForm.getUsername());
        requestParameters.put("password", loginForm.getPassword());
        TokenRequest tokenRequest = new TokenRequest(
                requestParameters,
                clientDetails.getClientId(), clientDetails.getScope(),
                requestParameters.get("grant_type")
        );
        ResourceOwnerPasswordTokenGranter granter = ApplicationContextUtil.getBean(ResourceOwnerPasswordTokenGranter.class);
        OAuth2AccessToken oAuth2AccessToken = granter.grant(tokenRequest.getGrantType(), tokenRequest);
        return new AccessToken(oAuth2AccessToken);
    }

    @Override
    public void logout() {
        try {
            String accessToken = SecurityContextUtil.getAccessToken();
            if (StringUtils.isNotBlank(accessToken)) {
                OAuth2AccessToken oAuth2AccessToken = tokenStore.readAccessToken(SecurityContextUtil.getAccessToken());
                tokenStore.removeAccessToken(oAuth2AccessToken);
                if (Objects.nonNull(oAuth2AccessToken.getRefreshToken())) {
                    tokenStore.removeRefreshToken(oAuth2AccessToken.getRefreshToken());
                }
                SecurityContextHolder.getContext().setAuthentication(null);
                SecurityContextHolder.clearContext();
            }
        } catch (Exception ignored) {}
    }

    @Override
    public User info() {
        String userId = SecurityContextUtil.getUserId();
        Optional<User> optional = Optional.ofNullable(authDao.queryUserByUserId(userId));
        return optional.orElse(null);
    }

    @Override
    public Permission permissions() {
        String userId = SecurityContextUtil.getUserId();
        List<Menu> menus = "admin".equals(userId) ?
                authDao.getMenuAll() :
                authDao.getMenuByUserId(userId);
        // 按钮权限
        Set<String> buttonList = menus.stream().filter(b -> b.getType() == 2)
                .map(Menu::getPermission)
                .collect(Collectors.toSet());
        // 构建菜单
        List<Menu> menuList = menus.stream().filter(m -> m.getType() != 2).collect(Collectors.toList());
        List<Route> routeList = this.buildMenu(menuList);
        // 权限信息
        Permission permission = new Permission();
        permission.setRoutes(routeList);
        permission.setButtons(buttonList);
        return permission;
    }

    @Override
    public List<Route> buildMenu(List<Menu> menus) {
        return menus.stream()
                .filter(m -> m.getPid() == 0)
                .sorted(Comparator.comparingInt(Menu::getSort))
                .map(m -> {
                    Route route = new Route(m);
                    route.setChildren(buildMenu(m, menus));
                    return route;
                }).collect(Collectors.toList());
    }

    private List<Route> buildMenu(Menu root, List<Menu> allListTree) {
        return allListTree.stream().filter((treeEntity) -> treeEntity.getPid().equals(root.getId()))
                .sorted(Comparator.comparingInt(Menu::getSort))
                .map((treeEntity) -> {
                    Route routeDTO = new Route(treeEntity);
                    routeDTO.setChildren(this.buildMenu(treeEntity, allListTree));
                    return routeDTO;
                }).collect(Collectors.toList());
    }
}
