package com.flow.service.impl;

import com.flow.common.core.exception.BaseException;
import com.flow.common.mybatis.service.impl.BaseServiceImpl;
import com.flow.common.oauth2.utils.SecurityContextUtil;
import com.flow.dao.FlowDefineDao;
import com.flow.entity.FlowDefine;
import com.flow.entity.node.*;
import com.flow.enums.ProcessStatus;
import com.flow.model.FormInfo;
import com.flow.model.StartProcess;
import com.flow.service.FlowDefineService;
import org.flowable.common.engine.impl.identity.Authentication;
import org.flowable.engine.RepositoryService;
import org.flowable.engine.RuntimeService;
import org.flowable.engine.repository.ProcessDefinition;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.CachePut;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.cache.annotation.Caching;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.Calendar;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

@Service
public class FlowDefineServiceImpl extends BaseServiceImpl<FlowDefineDao, FlowDefine> implements FlowDefineService {
    @Autowired
    private FlowDefineDao flowDefineDao;
    @Autowired
    private RepositoryService repositoryService;
    @Autowired
    private RuntimeService runtimeService;
    @Autowired
    private FlowDefineService _this;
    @Autowired
    private StringRedisTemplate stringRedisTemplate;

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void startProcess(StartProcess startProcess) {
        FlowDefine flowDefine = _this.getDefine(startProcess.getDefineId());
        if (Objects.isNull(flowDefine)) {
            throw new BaseException("流程不存在");
        }
        Integer suspend = flowDefine.getSuspend();
        if (suspend == 2) {
            throw new BaseException(String.format("流程（%s）已挂起，无法发起流程！", flowDefine.getName()));
        }
        Map<String, Object> values = startProcess.getValues();
        List<ApprovalNode> approvalNodes = flowDefine.findFlowNodesOfType(ApprovalNode.class);
        for (ApprovalNode approvalNode : approvalNodes) {
            values.put(String.format("%sCollection", approvalNode.getId()), approvalNode.getAssignees(values, flowDefine));
        }
        List<CcNode> ccNodes = flowDefine.findFlowNodesOfType(CcNode.class);
        for (CcNode ccNode : ccNodes) {
            values.put(String.format("%sCollection", ccNode.getId()), ccNode.getAssignees(values));
        }
        List<NotifyNode> notifyNodes = flowDefine.findFlowNodesOfType(NotifyNode.class);
        for (NotifyNode notifyNode : notifyNodes) {
            values.put(String.format("%sCollection", notifyNode.getId()), notifyNode.getAssignees(values));
        }
        // 生成流程实例ID
        String serialKey = "flow:serial";
        Long increment = stringRedisTemplate.opsForValue().increment(serialKey, 1);
        if (Objects.nonNull(increment) && increment.intValue() == 1) {
            // 设置缓存过期时间
            Calendar tomorrowMidnight = Calendar.getInstance();
            tomorrowMidnight.add(Calendar.DAY_OF_YEAR, 1);
            tomorrowMidnight.set(Calendar.HOUR_OF_DAY, 0);
            tomorrowMidnight.set(Calendar.MINUTE, 0);
            tomorrowMidnight.set(Calendar.SECOND, 0);
            tomorrowMidnight.set(Calendar.MILLISECOND, 0);
            long midnightTimestamp = tomorrowMidnight.getTimeInMillis() / 1000;
            stringRedisTemplate.expire(serialKey, midnightTimestamp, TimeUnit.SECONDS);
        }
        String serial = String.format("%3s", increment).replace(" ", "0");
        String currentDate = DateTimeFormatter.ofPattern("yyyyMMdd").format(LocalDate.now());
        String instanceId = String.format("fw%s%s", currentDate, serial);

        Authentication.setAuthenticatedUserId(SecurityContextUtil.getUserId());
        runtimeService.createProcessInstanceBuilder()
                .processDefinitionId(startProcess.getDefineId())
                .name(startProcess.getInstanceName())
                .owner(SecurityContextUtil.getUserId())
                .predefineProcessInstanceId(instanceId)
                .businessStatus(ProcessStatus.RUNNING.toString())
                .variables(values)
                .start();
    }

    @Override
    public List<FlowDefine> list(FlowDefine flowDefine) {
        return flowDefineDao.getList(flowDefine);
    }

    @Caching(
            put = @CachePut(value = "process", key = "#defineId", unless = "#result == null"),
            evict = @CacheEvict(value = "process", key = "#defineId", condition = "#result == null")
    )
    @Override
    public FlowDefine suspend(String defineId) {
        ProcessDefinition processDefinition = this.repositoryService.getProcessDefinition(String.valueOf(defineId));
        if (processDefinition.isSuspended()) {
            this.repositoryService.activateProcessDefinitionById(String.valueOf(defineId));
            return _this.getDefine(defineId);
        } else {
            this.repositoryService.suspendProcessDefinitionById(String.valueOf(defineId));
            return null;
        }
    }

    @Cacheable(key = "#defineId", value = "process", unless = "#result == null")
    @Override
    public FlowDefine getDefine(String defineId) {
        return flowDefineDao.getByDefineId(defineId);
    }

    @Override
    public FormInfo getStartForm(String defineId) {
        FlowDefine define = _this.getDefine(defineId);
        if (Objects.isNull(define)) {
            throw new BaseException("流程不存在");
        }
        StartNode startNode = (StartNode) define.getProcess();
        List<FormProperty> formProperties = startNode.getFormProperties();
        return new FormInfo(define, formProperties, null);
    }
}
