package com.flow.service.impl;

import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.flow.common.core.exception.BaseException;
import com.flow.common.mybatis.constant.SqlConstant;
import com.flow.common.mybatis.service.impl.BaseServiceImpl;
import com.flow.dao.FlowModelDao;
import com.flow.entity.FlowDefine;
import com.flow.entity.FlowModel;
import com.flow.service.FlowDefineService;
import com.flow.service.FlowModelHistoryService;
import com.flow.service.FlowModelService;
import org.flowable.bpmn.converter.BpmnXMLConverter;
import org.flowable.bpmn.model.BpmnModel;
import org.flowable.engine.RepositoryService;
import org.flowable.engine.repository.Deployment;
import org.flowable.engine.repository.ProcessDefinition;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletResponse;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URLEncoder;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

@Service
public class FlowModelServiceImpl extends BaseServiceImpl<FlowModelDao, FlowModel> implements FlowModelService {
    @Autowired
    private FlowModelDao flowModelDao;
    @Autowired
    private FlowModelHistoryService flowModelHistoryService;
    @Autowired
    private RepositoryService repositoryService;
    @Autowired
    private FlowDefineService flowDefineService;

    @Override
    public List<FlowModel> list(FlowModel flowModel) {
        return flowModelDao.lambdaQueryChain()
                .select(FlowModel::getId,
                        FlowModel::getCode,
                        FlowModel::getName,
                        FlowModel::getIcon,
                        FlowModel::getVersion,
                        FlowModel::getSort,
                        FlowModel::getGroupId,
                        FlowModel::getEnable,
                        FlowModel::getRemark,
                        FlowModel::getAdmin,
                        FlowModel::getCreatedBy,
                        FlowModel::getUpdatedBy,
                        FlowModel::getCreateTime,
                        FlowModel::getUpdateTime)
                .list();
    }

    @Override
    public void downloadXml(FlowModel flowModel) throws IOException {
        BpmnModel bpmnModel = flowModel.toBpmnModel();
        byte[] xmlBytes = new BpmnXMLConverter().convertToXML(bpmnModel);
        if (Objects.nonNull(xmlBytes)) {
            String fileName = flowModel.getName().replaceAll(" ", "_") + ".bpmn20.xml";
            HttpServletResponse response = ((ServletRequestAttributes) RequestContextHolder.getRequestAttributes()).getResponse();
            if (Objects.nonNull(response)) {
                response.setContentType("application/xml");
                response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
                response.setHeader("Content-Disposition", "attachment; filename=file.bpmn20.xml; filename*=" + URLEncoder.encode(fileName, "UTF-8"));
                ServletOutputStream servletOutputStream = response.getOutputStream();
                BufferedInputStream in = new BufferedInputStream(new ByteArrayInputStream(xmlBytes));
                byte[] buffer = new byte[8096];
                while (true) {
                    int count = in.read(buffer);
                    if (count == -1) {
                        break;
                    }
                    servletOutputStream.write(buffer, 0, count);
                }
                // 刷新并关闭流
                servletOutputStream.flush();
                servletOutputStream.close();
            }
        }
    }

    @Override
    public FlowModel getModel(Long modelId) {
        Optional<FlowModel> optional = flowModelDao.lambdaQueryChain()
                .eq(FlowModel::getId, modelId)
                .last(SqlConstant.LIMIT1)
                .oneOpt();
        return optional.orElse(null);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowModel create(FlowModel flowModel) {
        flowModel.setCode(String.format("flow-%s", IdWorker.getIdStr()));
        flowModelDao.insert(flowModel);
        return flowModel;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowModel update(FlowModel flowModel) {
        flowModelDao.updateById(flowModel);
        return flowModel;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void delete(Set<Long> ids) {
        flowModelDao.deleteBatchIds(ids);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowDefine deploy(Long modelId) {
        FlowModel flowModel = flowModelDao.selectById(modelId);
        if (!flowModel.getEnable()) {
            throw new BaseException("模型已停用无法发布");
        }
        // 保存历史
        flowModelHistoryService.save(flowModel);
        flowModel.setVersion(flowModel.getVersion() + 1);
        flowModelDao.updateById(flowModel);
        BpmnModel bpmnModel = flowModel.toBpmnModel();
        Deployment deploy = repositoryService.createDeployment()
                .key(flowModel.getCode())
                .name(flowModel.getName())
                .category(flowModel.getGroupId().toString())
                .addBpmnModel(String.format("%s.bpmn20.xml", flowModel.getName()), bpmnModel)
                .deploy();

        ProcessDefinition definition = repositoryService.createProcessDefinitionQuery()
                .deploymentId(deploy.getId())
                .singleResult();
        repositoryService.setProcessDefinitionCategory(definition.getId(), flowModel.getGroupId().toString());

        FlowDefine flowDefine = new FlowDefine(definition);
        flowDefine.setModelId(flowModel.getId());
        flowDefine.setProcess(flowModel.getProcess());
        flowDefine.setForm(flowModel.getForm());
        flowDefine.setIcon(flowModel.getIcon());
        flowDefine.setAdmin(flowModel.getAdmin());
        flowDefine.setSettings(flowModel.getSettings());
        flowDefineService.save(flowDefine);
        return flowDefine;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowModel suspend(Long modelId) {
        FlowModel flowModel = flowModelDao.selectById(modelId);
        boolean enable = !flowModel.getEnable();
        flowModel.setEnable(enable);
        flowModelDao.updateById(flowModel);
        ProcessDefinition processDefinition = repositoryService.createProcessDefinitionQuery()
                .processDefinitionKey(flowModel.getCode())
                .latestVersion()
                .singleResult();
        if (Objects.nonNull(processDefinition)) {
            flowDefineService.suspend(processDefinition.getId());
        }
        return flowModel;
    }
}
