package com.flow.service.impl;

import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.CollectionUtils;
import com.baomidou.mybatisplus.core.toolkit.StringUtils;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.flow.cmd.JumpActivityCmd;
import com.flow.common.core.exception.BaseException;
import com.flow.common.core.model.PageResult;
import com.flow.common.mybatis.constant.SqlConstant;
import com.flow.common.mybatis.service.impl.BaseServiceImpl;
import com.flow.common.oauth2.utils.SecurityContextUtil;
import com.flow.dao.FlowTaskDao;
import com.flow.entity.FlowActivity;
import com.flow.entity.FlowDefine;
import com.flow.entity.FlowInstance;
import com.flow.entity.FlowTask;
import com.flow.entity.node.ApprovalNode;
import com.flow.entity.node.Node;
import com.flow.enums.ProcessStatus;
import com.flow.model.*;
import com.flow.service.FlowActivityService;
import com.flow.service.FlowDefineService;
import com.flow.service.FlowInstanceService;
import com.flow.service.FlowTaskService;
import com.flow.utils.ProcessUtil;
import com.google.common.collect.Lists;
import org.flowable.bpmn.model.BpmnModel;
import org.flowable.bpmn.model.UserTask;
import org.flowable.engine.*;
import org.flowable.engine.runtime.ActivityInstance;
import org.flowable.task.api.DelegationState;
import org.flowable.task.api.Task;
import org.flowable.task.api.TaskInfo;
import org.flowable.task.api.history.HistoricTaskInstance;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class FlowTaskServiceImpl extends BaseServiceImpl<FlowTaskDao, FlowTask> implements FlowTaskService {
    @Autowired
    private FlowTaskDao flowTaskDao;
    @Autowired
    private TaskService taskService;
    @Autowired
    private RuntimeService runtimeService;
    @Autowired
    private HistoryService historyService;
    @Autowired
    private FlowDefineService flowDefineService;
    @Autowired
    private FlowInstanceService flowInstanceService;
    @Autowired
    private FlowActivityService flowActivityService;
    @Autowired
    private ManagementService managementService;
    @Autowired
    private RepositoryService repositoryService;

    @Override
    public PageResult<FlowTask> todo(FlowTaskQuery flowTaskQuery) {
        flowTaskQuery.setAssignee(SecurityContextUtil.getUserId());
        Page<FlowInstance> page = new Page<>(
                flowTaskQuery.getPage(),
                flowTaskQuery.getLimit()
        );
        IPage<FlowTask> iPage = flowTaskDao.todo(page, flowTaskQuery);
        return new PageResult<>(iPage.getTotal(), iPage.getRecords());
    }

    @Override
    public PageResult<FlowTask> done(FlowTaskQuery flowTaskQuery) {
        flowTaskQuery.setAssignee(SecurityContextUtil.getUserId());
        Page<FlowInstance> page = new Page<>(
                flowTaskQuery.getPage(),
                flowTaskQuery.getLimit()
        );
        IPage<FlowTask> iPage = flowTaskDao.done(page, flowTaskQuery);
        return new PageResult<>(iPage.getTotal(), iPage.getRecords());
    }

    @Override
    public FormInfo getForm(String taskId) {
        TaskInfo task = this.getTask(taskId);
        if (Objects.isNull(task)) {
            throw new BaseException("任务不存在");
        }
        FlowDefine flowDefine = flowDefineService.getDefine(task.getProcessDefinitionId());
        if (Objects.isNull(flowDefine)) {
            throw new BaseException("流程不存在");
        }
        Map<String, Object> variables = this.getVariables(task.getId());
        ApprovalNode node = flowDefine.getNode(task.getTaskDefinitionKey(), ApprovalNode.class);
        FormInfo formInfo = new FormInfo(
                flowDefine,
                node.getFormProperties(),
                variables,
                node.getOperations()
        );
        FlowInstance instance = flowInstanceService.getInstance(task.getProcessInstanceId());
        formInfo.setInstanceId(task.getProcessInstanceId());
        formInfo.setTaskId(task.getId());
        formInfo.setFormName(instance.getName());
        formInfo.setStartUserId(instance.getStartUserId());
        formInfo.setStartTime(instance.getStartTime());
        if (task instanceof HistoricTaskInstance) {
            HistoricTaskInstance historicTaskInstance = (HistoricTaskInstance) task;
            Date endTime = historicTaskInstance.getEndTime();
            if (Objects.nonNull(endTime)) {
                formInfo.setEndTime(
                        LocalDateTime.ofInstant(
                                endTime.toInstant(),
                                TimeZone.getDefault().toZoneId()
                        )
                );
            }
        }
        formInfo.setStatus(instance.getStatus());
        return formInfo;
    }

    @Override
    public TaskInfo getTask(String taskId) {
        Task task = taskService.createTaskQuery().taskId(taskId).singleResult();
        if (Objects.nonNull(task)) {
            return task;
        }
        return historyService.createHistoricTaskInstanceQuery().taskId(taskId).singleResult();
    }

    @Override
    public Map<String, Object> getVariables(String taskId) {
        TaskInfo task = this.getTask(taskId);
        if (Objects.isNull(task)) {
            throw new BaseException("任务不存在");
        }
        Map<String, Object> variables = new HashMap<>();
        if (task instanceof HistoricTaskInstance) {
            historyService.createHistoricVariableInstanceQuery().taskId(taskId).list().forEach(variableInstance ->
                    variables.putIfAbsent(variableInstance.getVariableName(), variableInstance.getValue())
            );
            historyService.createHistoricVariableInstanceQuery().processInstanceId(task.getProcessInstanceId()).list().forEach(variableInstance ->
                    variables.putIfAbsent(variableInstance.getVariableName(), variableInstance.getValue())
            );
        } else {
            taskService.getVariableInstances(taskId).values().forEach(variableInstance ->
                    variables.putIfAbsent(variableInstance.getName(), variableInstance.getValue())
            );
            runtimeService.getVariableInstances(task.getProcessInstanceId()).values().forEach(variableInstance ->
                    variables.putIfAbsent(variableInstance.getName(), variableInstance.getValue())
            );
        }
        Optional<FlowTask> optional = flowTaskDao.lambdaQueryChain()
                .select(FlowTask::getVariables)
                .eq(FlowTask::getId, taskId)
                .oneOpt();
        optional.ifPresent((flowTask) -> variables.putAll(flowTask.getVariables()));
        return variables;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void saveForm(TaskForm taskForm) {
        taskService.setVariables(taskForm.getTaskId(), taskForm.getValues());
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void reject(RejectTaskForm taskForm) {
        Task task = this.comment(taskForm);
        this.updateStatus(task.getId(), ProcessStatus.REJECT);
        flowActivityService.lambdaUpdate()
                .eq(FlowActivity::getTaskId, taskForm.getTaskId())
                .set(FlowActivity::getStatus, ProcessStatus.REJECT)
                .set(FlowActivity::getUserId, task.getAssignee())
                .update();
        // 回退节点
        List<String> executionIds = managementService.executeCommand(new JumpActivityCmd(
                task.getProcessInstanceId(),
                taskForm.getTargetNode()
        ));
        // 删除进行中节点
        if (CollectionUtils.isNotEmpty(executionIds)) {
            flowActivityService.lambdaUpdate()
                    .eq(FlowActivity::getInstanceId, task.getProcessInstanceId())
                    .eq(FlowActivity::getStatus, ProcessStatus.RUNNING)
                    .in(FlowActivity::getParentId, executionIds)
                    .remove();
        }
        // 操作记录
        Optional<FlowActivity> optional = flowActivityService.lambdaQuery()
                .eq(FlowActivity::getTaskId, taskForm.getTaskId())
                .oneOpt();
        if (optional.isPresent()) {
            FlowDefine define = flowDefineService.getDefine(task.getProcessDefinitionId());
            ApprovalNode approvalNode = define.getNode(taskForm.getTargetNode(), ApprovalNode.class);
            FlowActivity flowActivity = optional.get();
            List<Comment> comments = Optional.ofNullable(flowActivity.getComment()).orElse(Lists.newArrayList());
            Comment comment = new Comment();
            comment.setUserId(SecurityContextUtil.getUserId());
            comment.setTime(LocalDateTime.now());
            comment.setContent(String.format("<span>退回至 <span class='color-error'>%s</span></span>", approvalNode.getName()));
            comment.setAttachments(Lists.newArrayList());
            comments.add(comment);
            flowActivity.setComment(comments);
            flowActivityService.updateById(flowActivity);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void refuse(TaskForm taskForm) {
        Task task = this.comment(taskForm);
        this.updateStatus(task.getId(), ProcessStatus.REFUSE);
        flowActivityService.lambdaUpdate()
                .eq(FlowActivity::getTaskId, taskForm.getTaskId())
                .set(FlowActivity::getStatus, ProcessStatus.REFUSE)
                .set(FlowActivity::getUserId, task.getAssignee())
                .update();
        flowInstanceService.reject(task.getProcessInstanceId());
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void autoRefuse(String taskId) {
        this.autoRefuse(taskId, null);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void autoRefuse(String taskId, String content) {
        TaskInfo task = this.getTask(taskId);
        if (Objects.isNull(task)) {
            throw new BaseException("任务不存在");
        }
        Map<String, Object> variables = taskService.getVariables(task.getId());
        TaskForm taskForm = new TaskForm();
        taskForm.setTaskId(task.getId());
        taskForm.setValues(variables);
        Comment comment = new Comment();
        comment.setUserId("admin");
        comment.setTime(LocalDateTime.now());
        comment.setContent(String.format("<span><span class='color-error'>自动拒绝</span>%s</span>", Optional.ofNullable(content).orElse("")));
        taskForm.setComment(comment);
        this.refuse(taskForm);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void transfer(TransferTaskForm taskForm) {
        Task task = this.comment(taskForm);
        taskService.setAssignee(task.getId(), taskForm.getAssignee());
        Optional<FlowActivity> optional = flowActivityService.lambdaQuery()
                .eq(FlowActivity::getTaskId, taskForm.getTaskId())
                .last(SqlConstant.LIMIT1)
                .oneOpt();
        if (optional.isPresent()) {
            FlowActivity flowActivity = optional.get();
            flowActivity.setUserId(taskForm.getAssignee());
            flowActivityService.updateById(flowActivity);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void addSign(AddSignTaskForm taskForm) {
        Task taskInfo = this.comment(taskForm);
        String variableName = String.format("%sCollection", taskInfo.getTaskDefinitionKey());
        List<String> collectionValue = runtimeService.getVariable(
                taskInfo.getProcessInstanceId(),
                variableName,
                ArrayList.class
        );
        if (Objects.nonNull(collectionValue) && collectionValue.contains(taskForm.getAssignee())) {
            throw new BaseException("加签人员已存在");
        }
        int index = collectionValue.indexOf(taskInfo.getAssignee());
        if (index != -1) {
            collectionValue.add(index + 1, taskForm.getAssignee());
        } else {
            collectionValue.add(taskForm.getAssignee());
        }
        Map<String, Object> executionVariables = new HashMap<>();
        executionVariables.put(variableName, collectionValue);
        executionVariables.put("nrOfActiveInstances", collectionValue.size());
        runtimeService.setVariables(taskInfo.getProcessInstanceId(), executionVariables);
        executionVariables.clear();
        executionVariables.put(String.format("%sItem", taskInfo.getTaskDefinitionKey()), taskForm.getAssignee());
        runtimeService.addMultiInstanceExecution(taskInfo.getTaskDefinitionKey(), taskInfo.getProcessInstanceId(), executionVariables);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void removeSign(AddSignTaskForm taskForm) {
        TaskInfo taskInfo = this.getTask(taskForm.getTaskId());
        List<Task> tasks = taskService.createTaskQuery()
                .processInstanceId(taskInfo.getProcessInstanceId())
                .taskDefinitionKey(taskInfo.getTaskDefinitionKey())
                .taskAssignee(taskForm.getAssignee())
                .list();
        for (Task task : tasks) {
            String executionId = task.getExecutionId();
            runtimeService.deleteMultiInstanceExecution(executionId, false);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void delegate(TransferTaskForm taskForm) {
        Task task = this.comment(taskForm);
        taskService.delegateTask(task.getId(), taskForm.getAssignee());
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void updateStatus(String id, ProcessStatus status) {
        this.updateStatus(id, status, null);
    }

    @Override
    public void updateStatus(String id, ProcessStatus status, Map<String, Object> variables) {
        FlowTask flowTask = new FlowTask();
        flowTask.setId(id);
        flowTask.setStatus(status);
        flowTask.setVariables(variables);
        this.updateById(flowTask);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void complete(TaskForm taskForm) {
        Task task = this.comment(taskForm);
        DelegationState delegationState = task.getDelegationState();
        Map<String, Object> transientVariables = new HashMap<>();
        Map<String, Object> variables = taskService.getVariables(task.getId());
        Map<String, Object> values = taskForm.getValues();
        if (CollectionUtils.isNotEmpty(values)) {
            values.forEach((key, value) -> {
                if (variables.containsKey(key)) {
                    Object obj = variables.get(key);
                    if (Objects.nonNull(obj) && !obj.equals(value)) {
                        transientVariables.put(key, value);
                    }
                }
            });
        }
        if (Objects.nonNull(delegationState) && delegationState == DelegationState.PENDING) {
            taskService.resolveTask(
                    taskForm.getTaskId(),
                    values
            );
        } else {
            taskService.complete(
                    taskForm.getTaskId(),
                    values
            );
            this.updateStatus(
                    taskForm.getTaskId(),
                    ProcessStatus.COMPLETED,
                    transientVariables
            );
        }
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void autoComplete(String taskId) {
        this.autoComplete(taskId, null);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void autoComplete(String taskId, String content) {
        TaskInfo task = this.getTask(taskId);
        if (Objects.isNull(task)) {
            throw new BaseException("任务不存在");
        }
        Map<String, Object> variables = taskService.getVariables(task.getId());
        TaskForm taskForm = new TaskForm();
        taskForm.setTaskId(task.getId());
        taskForm.setValues(variables);
        Comment comment = new Comment();
        comment.setUserId("admin");
        comment.setTime(LocalDateTime.now());
        comment.setContent(String.format("<span><span class='color-primary'>自动通过</span>%s</span>", Optional.ofNullable(content).orElse("")));
        taskForm.setComment(comment);
        this.complete(taskForm);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public Task comment(TaskForm taskForm) {
        Task task = taskService.createTaskQuery().taskId(taskForm.getTaskId()).singleResult();
        if (Objects.isNull(task)) {
            throw new BaseException("任务不存在");
        }
        if (task.isSuspended()) {
            throw new BaseException("任务已挂起");
        }
        Comment comment = taskForm.getComment();
        if (Objects.nonNull(comment)) {
            if (StringUtils.isNotBlank(comment.getContent()) || CollectionUtils.isNotEmpty(comment.getAttachments())) {
                comment.setUserId(SecurityContextUtil.getUserId());
                comment.setTime(LocalDateTime.now());
                Optional<FlowActivity> optional = flowActivityService.lambdaQuery()
                        .eq(FlowActivity::getTaskId, task.getId())
                        .last("limit 1")
                        .oneOpt();
                if (optional.isPresent()) {
                    FlowActivity flowActivity = optional.get();
                    List<Comment> comments = Optional.ofNullable(flowActivity.getComment()).orElse(Lists.newArrayList());
                    comments.add(comment);
                    flowActivity.setComment(comments);
                    flowActivityService.updateById(flowActivity);
                }
            }
        }
        return task;
    }

    @Override
    public List<Node> getParentNodes(String taskId) {
        TaskInfo taskInfo = this.getTask(taskId);
        if (Objects.isNull(taskInfo)) {
            throw new BaseException("任务不存在");
        }
        BpmnModel model = repositoryService.getBpmnModel(taskInfo.getProcessDefinitionId());
        Set<UserTask> parentElements = ProcessUtil.findParentElements(model.getFlowElement(taskInfo.getTaskDefinitionKey()), UserTask.class);
        String sql = "SELECT t1.ID_,t1.ACT_ID_,t1.ACT_NAME_ FROM ACT_RU_ACTINST AS t1 WHERE " +
                "END_TIME_ = ( SELECT MAX( END_TIME_ ) FROM ACT_RU_ACTINST AS t2 WHERE t1.ACT_ID_ = t2.ACT_ID_ ) " +
                "AND t1.PROC_INST_ID_ = #{processInstanceId} " +
                "AND t1.ACT_TYPE_ = 'userTask' " +
                "AND t1.END_TIME_ IS NOT NULL " +
                "ORDER BY t1.END_TIME_ DESC";
        List<ActivityInstance> activityInstances = runtimeService.createNativeActivityInstanceQuery()
                .sql(sql)
                .parameter("processInstanceId", taskInfo.getProcessInstanceId())
                .list();
        return activityInstances.stream()
                .filter(activityInstance -> parentElements.stream().anyMatch(userTask -> userTask.getId().equals(activityInstance.getActivityId())))
                .map(activity -> {
                    Node node = new ApprovalNode();
                    node.setId(activity.getActivityId());
                    node.setName(activity.getActivityName());
                    return node;
                })
                .collect(Collectors.toList());
    }

    @Override
    public List<String> getMultiInstanceUnStartedUsers(String taskId) {
        Task task = taskService.createTaskQuery()
                .taskId(taskId)
                .singleResult();
        List<Task> list = taskService.createTaskQuery()
                .processInstanceId(task.getProcessInstanceId())
                .taskDefinitionKey(task.getTaskDefinitionKey())
                .list();
        return list.stream().map(Task::getAssignee).collect(Collectors.toList());
    }
}
